import re

from pathlib import Path
from .typings import PathLike, AnalysisResult


def compact_json_number_array(json_str: str):
  """Compact numbers (including floats) in JSON arrays to be on the same line."""
  return re.sub(
    r'(\[\n(?:\s*\d+(\.\d+)?,\n)+\s*\d+(\.\d+)?\n\s*\])',
    lambda m: m.group(1).replace('\n', '').replace(' ', ''),
    json_str
  )


def mkpath(path: PathLike):
  return Path(path).expanduser().resolve()


def load_result(
  path: PathLike,
  load_activations: bool = True,
  load_embeddings: bool = True,
) -> AnalysisResult:
  path = mkpath(path)
  result = AnalysisResult.from_json(
    path,
    load_activations=load_activations,
    load_embeddings=load_embeddings,
  )
  return result



import numpy as np
from scipy.ndimage import maximum_filter1d


def widen_temporal_events(events, num_neighbors=2):
  """Widen temporal events by a given number of neighbors."""
  widen_events = events
  for i in range(num_neighbors):
    widen_events = maximum_filter1d(widen_events, size=3)
    neighbor_indices = np.flatnonzero((events != 1) & (widen_events > 0))
    widen_events[neighbor_indices] *= 0.5
  
  return widen_events



import numpy as np
import torch
from collections import defaultdict


def collate_fn(raw_batch):
  variable_length_batch = defaultdict(list)
  for row in raw_batch:
    for key, value in list(row.items()):
      if isinstance(value, list):
        variable_length_batch[key].append(row.pop(key))
  
  max_T = max(x['spec'].shape[1] for x in raw_batch)
  batch = []
  for raw_data in raw_batch:
    data = {}
    for key, value in raw_data.items():
      if key in ['track_key', 'true_bpm', 'widen_true_bpm', 'true_bpm_int']:
        data[key] = value
      elif key in [
        'true_beat', 'true_downbeat', 'true_section', 'true_function',
        'widen_true_beat', 'widen_true_downbeat', 'widen_true_section',
      ]:
        data[key] = value[:max_T]
      elif key in ['spec']:
        T = raw_data[key].shape[1]
        spec = raw_data[key]
        if T < max_T:
          spec = np.pad(spec, ((0, 0), (0, max_T - T), (0, 0)), 'constant')
          mask = np.pad(np.ones(T), (0, max_T - T), 'constant')
        else:
          mask = np.ones(max_T)
        data[key] = spec
        data['mask'] = mask
      else:
        raise ValueError(f'Unknown key: {key}')
    batch.append(data)
  
  batch = torch.utils.data.default_collate(batch)
  batch = {**batch, **variable_length_batch}
  
  return batch



import numpy as np

import os
import wandb
import mir_eval

from pprint import pprint
from typing import List, Tuple, Dict, Mapping
from functools import partial
from tqdm import tqdm
from lightning import Trainer
from torch.multiprocessing import Pool
from madmom.evaluation.beats import BeatEvaluation

from .data import HarmonixDataModule
from .helpers import makeup_wandb_config
from .trainer import AllInOneTrainer
from .helpers import find_best_thresholds
from ..utils import mkpath
from ..config import Config, HARMONIX_LABELS
from ..typings import AllInOneOutput, AllInOnePrediction
from ..postprocessing import postprocess_functional_structure, postprocess_metrical_structure

DEBUG = False
RUN_ID = [
  'xxxxxxxx',  # Your W&B run ID here.
]
OUTDIR = 'eval/'


def main():
  global RUN_ID

  for run_id in tqdm(RUN_ID):
    print(f'=> Running evaluation of {run_id}...')
    evaluate(run_id=run_id)


def evaluate(
  run_id=None,
  model: AllInOneTrainer = None,
  trainer: Trainer = None,
):
  print('=> Evaluating...')
  if run_id:
    model, cfg, run = load_wandb_run(run_id, run_dir=OUTDIR)
    cfg.debug = DEBUG
  else:
    assert model is not None, 'Either run_id or model should be provided'
    assert trainer is not None, 'Trainer should be provided if model is provided'
    cfg = model.cfg
    run = wandb.run

  print('=> Creating data module...')
  if cfg.data.name == 'harmonix':
    dm = HarmonixDataModule(cfg)
  else:
    raise ValueError(f'Unknown dataset: {cfg.data.name}')

  if trainer is None:
    trainer = Trainer(
      # accelerator='cpu',
      devices=1,
    )

  if not cfg.debug:
    print('=> Finding best thresholds...')
    # Find the optimal thresholds.
    if (
      'best_threshold_beat' not in run.config
      or run.config['best_threshold_beat'] is None
    ):
      dm.setup('validate')
      outputs_val = trainer.predict(model, dataloaders=dm.val_dataloader())
      threshold_beat, threshold_downbeat = find_best_thresholds(outputs_val, cfg)

      if not cfg.debug:
        run.config.update({
          'best_threshold_beat': threshold_beat.item(),
          'best_threshold_downbeat': threshold_downbeat.item(),
        }, allow_val_change=True)
        if hasattr(run, 'update'):
          run.update()

    cfg.threshold_beat = run.config['best_threshold_beat']
    cfg.threshold_downbeat = run.config['best_threshold_downbeat']

  print(f'=> Evaluating with thresholds: {cfg.threshold_beat}, {cfg.threshold_downbeat}')

  scores = trainer.test(model, datamodule=dm)[0]
  if not cfg.debug:
    run.summary.update(scores)

  predict_outputs = trainer.predict(model, datamodule=dm)
  scores = compute_postprocessed_scores(predict_outputs, cfg, prefix='test/')

  print('=> Postprocessed scores on test set:')
  pprint(scores)
  if not cfg.debug:
    run.summary.update(scores)


def compute_postprocessed_scores(
  predict_outputs: List[Tuple[Dict, AllInOneOutput, AllInOnePrediction]],
  cfg: Config,
  prefix: str = '',
):
  all_scores: List[Mapping[str, float]] = []

  with Pool(os.cpu_count() // 2) as pool:
    fn = partial(
      compute_postprocessed_scores_step,
      cfg=cfg,
    )
    if cfg.debug:
      iterator = map(fn, predict_outputs)
    else:
      iterator = pool.imap(fn, predict_outputs)
    iterator = tqdm(iterator, total=len(predict_outputs), desc='Postprocessing...')

    for result in iterator:
      all_scores.append(result)

  avg_scores = {
    f'{prefix}{k}': np.mean([scores[k] for scores in all_scores])
    for k in all_scores[0].keys()
  }

  return avg_scores


def compute_postprocessed_scores_step(
  predict_output: Tuple[Dict, AllInOneOutput, AllInOnePrediction],
  cfg: Config,
) -> Mapping[str, float]:
  inputs, outputs, preds = predict_output

  pred_functional = postprocess_functional_structure(outputs, cfg)
  pred_metrical = postprocess_metrical_structure(outputs, cfg)

  eval_beat = BeatEvaluation(pred_metrical['beats'], inputs['true_beat_times'][0])
  eval_downbeat = BeatEvaluation(pred_metrical['downbeats'], inputs['true_downbeat_times'][0])

  scores_metrical = {
    'beat/f1': eval_beat.fmeasure,
    'beat/precision': eval_beat.precision,
    'beat/recall': eval_beat.recall,
    'beat/cmlt': eval_beat.cmlt,
    'beat/amlt': eval_beat.amlt,
    'downbeat/f1': eval_downbeat.fmeasure,
    'downbeat/precision': eval_downbeat.precision,
    'downbeat/recall': eval_downbeat.recall,
    'downbeat/cmlt': eval_downbeat.cmlt,
    'downbeat/amlt': eval_downbeat.amlt,
  }

  # Process ground truths.
  true_labels = inputs['true_function_list'][0]
  true_boundary_times = inputs['true_section_times'][0]
  # Some first positions are negative.
  true_boundary_times = np.maximum(true_boundary_times, 0)
  # Some last positions are out of audio length.
  duration = inputs['spec'].shape[2] * cfg.hop_size / cfg.sample_rate
  if true_boundary_times[-1] >= duration:
    true_boundary_times = true_boundary_times[:-1]
    true_labels = true_labels[:-1]
  if true_boundary_times[0] == 0:
    true_labels = true_labels[1:]  # there is no "start", beginning with "intro"
  else:
    # Else, insert "start" boundary at the beginning (0.0).
    true_boundary_times = np.insert(true_boundary_times, 0, 0.0)
  if true_boundary_times[-1] != duration:
    true_boundary_times = np.append(true_boundary_times, duration)

  pred_labels = [HARMONIX_LABELS.index(s.label) for s in pred_functional]
  pred_boundaries = np.array([[p.start, p.end] for p in pred_functional])
  true_boundaries = np.stack([true_boundary_times[:-1], true_boundary_times[1:]]).T

  scores_functional = mir_eval.segment.evaluate(
    true_boundaries, true_labels, pred_boundaries, pred_labels,
    trim=False
  )
  scores_functional = {f'segment/{k}': v for k, v in scores_functional.items()}

  scores = {**scores_functional, **scores_metrical}

  return scores


def load_wandb_run(
  run_id: str,
  run_dir: str = './wandb',
) -> Tuple[AllInOneTrainer, Config, wandb.apis.public.Run]:
  api = wandb.Api()
  run = api.run(f'taejun/danceformer/{run_id}')
  artifact = api.artifact(f'taejun/danceformer/model-{run_id}:latest', type='model')
  artifact_dir = artifact.download()
  checkpoint_path = mkpath(artifact_dir) / 'model.ckpt'
  outdir = mkpath(run_dir) / run_id
  outdir.mkdir(parents=True, exist_ok=True)
  cfg = makeup_wandb_config(run.config)
  model = AllInOneTrainer.load_from_checkpoint(
    checkpoint_path,
    map_location='cpu',
    cfg=cfg,
  )
  return model, cfg, run


if __name__ == '__main__':
  main()



import warnings
import librosa
import numpy as np
import torch.nn.functional as F
import torch

from typing import Dict, Union
from lightning import LightningModule
from madmom.evaluation.beats import BeatEvaluation, BeatMeanEvaluation
from numpy.typing import NDArray
from sklearn.metrics import f1_score, accuracy_score
from timm.optim.optim_factory import create_optimizer_v2 as create_optimizer
from timm.scheduler import create_scheduler
from timm.scheduler.scheduler import Scheduler

from ..models import AllInOne
from ..typings import AllInOneOutput, AllInOnePrediction
from ..config import Config
from .helpers import local_maxima

# For ignoring following warnings from madmom.evaluation
warnings.filterwarnings('ignore', category=RuntimeWarning, message='Mean of empty slice')
warnings.filterwarnings('ignore', category=UserWarning, message='Not enough beat annotations')
warnings.filterwarnings('ignore', category=UserWarning, message='The epoch parameter')
warnings.filterwarnings('ignore', category=UserWarning, message='no annotated tempo strengths given')


class AllInOneTrainer(LightningModule):
  scheduler: Scheduler

  def __init__(self, cfg: Config):
    super().__init__()
    self.cfg = cfg

    if cfg.model == 'allinone':
      self.model = AllInOne(cfg)
    else:
      raise NotImplementedError(f'Unknown model: {cfg.model}')

    self.lr = cfg.lr

  def forward(self, x):
    return self.model(x)

  def configure_optimizers(self):
    optimizer = create_optimizer(
      self,
      opt=self.cfg.optimizer,
      lr=self.cfg.lr,
      weight_decay=self.cfg.weight_decay,
    )
    if self.cfg.sched is not None:
      self.scheduler, _ = create_scheduler(self.cfg, optimizer)

    return {
      'optimizer': optimizer,
    }

  def on_train_epoch_end(self) -> None:
    if self.cfg.sanity_check:
      return

    if self.cfg.sched == 'plateau':
      if (self.current_epoch + 1) % self.cfg.validation_interval_epochs == 0:
        optimizer = self.trainer.optimizers[0]
        old_lr = optimizer.param_groups[0]['lr']

        metric = self.trainer.callback_metrics[self.cfg.eval_metric]
        self.scheduler.step(epoch=self.current_epoch + 1, metric=metric)

        new_lr = optimizer.param_groups[0]['lr']
        if new_lr < old_lr:
          print(f'=> The LR is decayed from {old_lr} to {new_lr}. '
                f'Loading the best model: {self.cfg.eval_metric}={self.trainer.checkpoint_callback.best_model_score}')
          self.load_from_checkpoint(self.trainer.checkpoint_callback.best_model_path, cfg=self.cfg)
      elif self.current_epoch + 1 <= self.cfg.warmup_epochs:
        self.scheduler.step(epoch=self.current_epoch + 1)
    else:
      self.scheduler.step(epoch=self.current_epoch + 1)

  def training_step(self, batch, batch_idx):
    batch_size = batch['spec'].shape[0]
    outputs: AllInOneOutput = self(batch['spec'])
    losses = self.compute_losses(outputs, batch, prefix='train/')
    loss = losses.pop('train/loss')
    self.log('train/loss', loss, prog_bar=True, batch_size=batch_size)
    self.log_dict(losses, batch_size=batch_size)

    if (self.current_epoch + 1) % self.cfg.validation_interval_epochs == 0 or self.cfg.debug:
      predictions = self.compute_predictions(outputs, mask=batch['mask'])
      scores = self.compute_metrics(predictions, batch, prefix='train/')
      self.log_dict(scores, sync_dist=True, on_epoch=True, batch_size=batch_size)

      if self.cfg.sanity_check:
        print('\n')
        for k, v in {**losses, **scores}.items():
          print(k, v.item())
        print('\n')

    return loss

  def evaluation_step(self, batch, batch_idx, prefix=None):
    batch_size = batch['spec'].shape[0]
    outputs: AllInOneOutput = self(batch['spec'])
    losses = self.compute_losses(outputs, batch, prefix)
    predictions = self.compute_predictions(outputs)
    scores = self.compute_metrics(predictions, batch, prefix)
    self.log_dict(losses, sync_dist=True, batch_size=batch_size)
    self.log_dict(scores, sync_dist=True, batch_size=batch_size)

  def validation_step(self, batch, batch_idx):
    self.evaluation_step(batch, batch_idx, prefix='val/')

  def test_step(self, batch, batch_idx):
    self.evaluation_step(batch, batch_idx, prefix='test/')

  def predict_step(self, batch, batch_idx: int, dataloader_idx: int = 0):
    assert batch['spec'].shape[0] == 1, 'Batch size must be 1 for prediction'
    outputs: AllInOneOutput = self(batch['spec'])
    # losses = self.compute_losses(outputs, batch)
    predictions = self.compute_predictions(outputs)
    # scores = self.compute_metrics(predictions, batch)
    return batch, outputs, predictions

  def compute_losses(self, outputs: AllInOneOutput, batch: Dict, prefix: str = None):
    loss = 0.0
    losses = {}

    loss_beat = F.binary_cross_entropy_with_logits(
      outputs.logits_beat, batch['widen_true_beat'],
      reduction='none',
    )
    loss_downbeat = F.binary_cross_entropy_with_logits(
      outputs.logits_downbeat, batch['widen_true_downbeat'],
      reduction='none',
    )
    loss_section = F.binary_cross_entropy_with_logits(
      outputs.logits_section, batch['widen_true_section'],
      reduction='none',
    )
    loss_function = F.cross_entropy(
      outputs.logits_function, batch['true_function'],
      reduction='none',
    )

    loss_beat = torch.mean(batch['mask'] * loss_beat)
    loss_downbeat = torch.mean(batch['mask'] * loss_downbeat)
    loss_section = torch.mean(batch['mask'] * loss_section)
    loss_function = torch.mean(batch['mask'] * loss_function)

    loss_beat *= self.cfg.loss_weight_beat
    loss_downbeat *= self.cfg.loss_weight_downbeat
    loss_section *= self.cfg.loss_weight_section
    loss_function *= self.cfg.loss_weight_function

    if self.cfg.learn_rhythm:
      loss += loss_beat + loss_downbeat
    if self.cfg.learn_structure:
      if self.cfg.learn_label:
        loss += loss_function
      if self.cfg.learn_segment:
        loss += loss_section

    losses.update(
      loss=loss,
      loss_beat=loss_beat,
      loss_downbeat=loss_downbeat,
      loss_section=loss_section,
      loss_function=loss_function,
    )
    if prefix:
      losses = prefix_dict(losses, prefix)
    return losses

  def compute_predictions(self, outputs: AllInOneOutput, mask=None):
    raw_prob_beats = torch.sigmoid(outputs.logits_beat.detach())
    raw_prob_downbeats = torch.sigmoid(outputs.logits_downbeat.detach())
    raw_prob_sections = torch.sigmoid(outputs.logits_section.detach())
    raw_prob_functions = torch.softmax(outputs.logits_function.detach(), dim=1)

    prob_beats, _ = local_maxima(raw_prob_beats, filter_size=self.cfg.min_hops_per_beat + 1)
    prob_downbeats, _ = local_maxima(raw_prob_downbeats, filter_size=4 * self.cfg.min_hops_per_beat + 1)
    prob_sections, _ = local_maxima(raw_prob_sections, filter_size=4 * self.cfg.min_hops_per_beat + 1)
    prob_functions = raw_prob_functions.cpu().numpy()

    if mask is not None:
      prob_beats *= mask
      prob_downbeats *= mask
      prob_sections *= mask

    pred_beats = prob_beats > self.cfg.threshold_beat
    pred_downbeats = prob_downbeats > self.cfg.threshold_downbeat
    pred_sections = prob_sections > self.cfg.threshold_section
    pred_functions = np.argmax(prob_functions, axis=1)
    if mask is not None:
      pred_functions = np.where(mask.cpu().numpy(), pred_functions, -1)

    pred_beat_times = self.tensor_to_time(pred_beats)
    pred_downbeat_times = self.tensor_to_time(pred_downbeats)
    pred_section_times = self.tensor_to_time(pred_sections)

    p = AllInOnePrediction(
      raw_prob_beats=raw_prob_beats,
      raw_prob_downbeats=raw_prob_downbeats,
      raw_prob_sections=raw_prob_sections,
      raw_prob_functions=raw_prob_functions,

      prob_beats=prob_beats,
      prob_downbeats=prob_downbeats,
      prob_sections=prob_sections,
      prob_functions=prob_functions,

      pred_beats=pred_beats,
      pred_downbeats=pred_downbeats,
      pred_sections=pred_sections,
      pred_functions=pred_functions,

      pred_beat_times=pred_beat_times,
      pred_downbeat_times=pred_downbeat_times,
      pred_section_times=pred_section_times,
    )

    return p

  def compute_metrics(self, p: AllInOnePrediction, batch: Dict, prefix: str = None):
    eval_beat = [
      BeatEvaluation(pred, true)
      for pred, true in zip(p.pred_beat_times, batch['true_beat_times'])
      if len(pred) > 1 and len(true) > 1
    ]
    eval_downbeat = [
      BeatEvaluation(pred, true)
      for pred, true in zip(p.pred_downbeat_times, batch['true_downbeat_times'])
      if len(pred) > 1 and len(true) > 1
    ]
    eval_section = [
      BeatEvaluation(pred, true, fmeasure_window=0.5)
      for pred, true in zip(p.pred_section_times, batch['true_section_times'])
      if len(pred) > 1 and len(true) > 1
    ]

    score_beat = BeatMeanEvaluation(eval_beat)
    score_downbeat = BeatMeanEvaluation(eval_downbeat)
    score_section = BeatMeanEvaluation(eval_section)

    mask = batch['mask'].cpu().numpy()
    true_functions = batch['true_function'].cpu().numpy()
    true_functions = np.where(mask, true_functions, -1)
    true_functions = true_functions.flatten()
    true_functions = true_functions[true_functions != -1]

    pred_functions = p.pred_functions
    pred_functions = np.where(mask, pred_functions, -1)
    pred_functions = pred_functions.flatten()
    pred_functions = pred_functions[pred_functions != -1]

    function_f1 = f1_score(true_functions, pred_functions, average='macro')
    function_accuracy = accuracy_score(true_functions, pred_functions)

    d = dict(
      beat_f1=score_beat.fmeasure,
      beat_precision=score_beat.precision,
      beat_recall=score_beat.recall,
      beat_cmlt=score_beat.cmlt,
      beat_amlt=score_beat.amlt,
      downbeat_f1=score_downbeat.fmeasure,
      downbeat_precision=score_downbeat.precision,
      downbeat_recall=score_downbeat.recall,
      downbeat_cmlt=score_downbeat.cmlt,
      downbeat_amlt=score_downbeat.amlt,
      section_f1=score_section.fmeasure,
      section_precision=score_section.precision,
      section_recall=score_section.recall,
      function_f1=function_f1,
      function_acc=function_accuracy,
    )

    if prefix:
      d = prefix_dict(d, prefix)
    return d

  def tensor_to_time(self, tensor: Union[torch.Tensor, NDArray]):
    """
    Args:
      tensor: a binary event tensor with shape (batch, frame)
    """
    if torch.is_tensor(tensor):
      tensor = tensor.cpu().numpy()
    batch_size = tensor.shape[0]
    i_examples, i_frames = np.where(tensor)
    times = librosa.frames_to_time(i_frames, sr=self.cfg.sample_rate, hop_length=self.cfg.hop_size)
    times = [times[i_examples == i] for i in range(batch_size)]
    return times

  def on_fit_end(self):
    print('=> Fit ended.')
    if self.trainer.is_global_zero and self.trainer.checkpoint_callback.best_model_path:
      print('=> Loading best model...')
      self.load_from_checkpoint(self.trainer.checkpoint_callback.best_model_path, cfg=self.cfg)
      print('=> Loaded best model.')


def prefix_dict(d: Dict, prefix: str):
  return {
    prefix + key: value
    for key, value in d.items()
  }



import numpy as np
import torch.nn.functional as F

import ast
import torch
import librosa

from typing import Union
from tqdm import tqdm
from omegaconf import DictConfig
from numpy.typing import NDArray
from madmom.evaluation.beats import BeatEvaluation, BeatMeanEvaluation
from ..config import Config
from ..typings import AllInOnePrediction


def makeup_wandb_config(run_config):
  run_config = run_config.copy()
  if 'data' in run_config:
    run_config['data'] = ast.literal_eval(run_config['data'])
  elif 'data/name' in run_config:
    run_config = unflatten_config(run_config, delimiter='/')
  return DictConfig(run_config)


def unflatten_config(d, delimiter='/'):
  unflattened = unflatten_dict(d, delimiter=delimiter)
  return DictConfig(unflattened)


def unflatten_dict(d, delimiter='/'):
  unflattened = {}
  for key, value in d.items():
    keys = key.split(delimiter)
    current_level = unflattened
    for k in keys[:-1]:
      current_level = current_level.setdefault(k, {})
    current_level[keys[-1]] = value
  return unflattened


def event_frames_to_time(
  tensor: Union[torch.Tensor, NDArray],
  cfg: Config = None,
  sample_rate: int = None,
  hop_size: int = None,
):
  """
  Args:
    tensor: a binary event tensor with shape (batch, frame)
  """
  assert len(tensor.shape) in (1, 2), 'Input tensor should have 1 or 2 dimensions'

  if cfg is not None:
    sample_rate = cfg.sample_rate
    hop_size = cfg.hop_size

  if torch.is_tensor(tensor):
    tensor = tensor.cpu().numpy()

  original_shape = tensor.shape
  if len(original_shape) == 1:
    tensor = tensor[None, :]

  batch_size = tensor.shape[0]
  i_examples, i_frames = np.where(tensor)
  times = librosa.frames_to_time(i_frames, sr=sample_rate, hop_length=hop_size)
  times = [times[i_examples == i] for i in range(batch_size)]

  if len(original_shape) == 1:
    times = times[0]
  return times


def local_maxima(tensor, filter_size=41):
  assert len(tensor.shape) in (1, 2), 'Input tensor should have 1 or 2 dimensions'
  assert filter_size % 2 == 1, 'Filter size should be an odd number'

  original_shape = tensor.shape
  if len(original_shape) == 1:
    tensor = tensor.unsqueeze(0)

  # Pad the input array with the minimum value
  padding = filter_size // 2
  padded_arr = F.pad(tensor, (padding, padding), mode='constant', value=-torch.inf)

  # Create a rolling window view of the padded array
  rolling_view = padded_arr.unfold(1, filter_size, 1)

  # Find the indices of the local maxima
  center = filter_size // 2
  local_maxima_mask = torch.eq(rolling_view[:, :, center], torch.max(rolling_view, dim=-1).values)
  local_maxima_indices = local_maxima_mask.nonzero()

  # Initialize a new PyTorch tensor with zeros and the same shape as the input tensor
  output_arr = torch.zeros_like(tensor)

  # Set the local maxima values in the output tensor
  output_arr[local_maxima_mask] = tensor[local_maxima_mask]

  output_arr = output_arr.reshape(original_shape)

  return output_arr, local_maxima_indices


def find_best_thresholds(predict_outputs, cfg: Config):
  probs_beat, trues_beat = [], []
  probs_downbeat, trues_downbeat = [], []
  for inputs, outputs, preds in predict_outputs:
    preds: AllInOnePrediction
    probs_beat.append(preds.raw_prob_beats[0])
    probs_downbeat.append(preds.raw_prob_downbeats[0])
    trues_beat.append(inputs['true_beat_times'][0])
    trues_downbeat.append(inputs['true_downbeat_times'][0])

  print('=> Start finding best thresholds for beat and downbeat...')
  threshold_beat, _ = find_best_threshold(probs_beat, trues_beat, cfg, cfg.min_hops_per_beat + 1)
  threshold_downbeat, _ = find_best_threshold(probs_downbeat, trues_downbeat, cfg, 4 * cfg.min_hops_per_beat + 1)

  return threshold_beat, threshold_downbeat


def find_best_threshold(probs, trues, cfg: Config, filter_size: int):
  results = {}
  for threshold in tqdm(
    np.linspace(0, 0.5, 51),
    desc='Finding best threshold...',
  ):
    results[threshold] = []
    for prob, true in zip(probs, trues):
      lmprob_beats, _ = local_maxima(prob, filter_size=filter_size)
      pred_beats = (lmprob_beats > threshold).numpy()
      pred_beat_times = event_frames_to_time(pred_beats, cfg)
      eval = BeatEvaluation(pred_beat_times, true)
      results[threshold].append(eval)
    results[threshold] = BeatMeanEvaluation(results[threshold])

  best_threshold = max(results, key=lambda k: results[k].fmeasure)
  return best_threshold, results[best_threshold]



import hydra
import lightning
from lightning import Trainer
from lightning.pytorch.callbacks import StochasticWeightAveraging, ModelCheckpoint, LearningRateMonitor, EarlyStopping
from lightning.pytorch.loggers import WandbLogger
from omegaconf import OmegaConf

from .data import HarmonixDataModule
from .trainer import AllInOneTrainer
from .evaluate import evaluate
from ..config import Config


@hydra.main(version_base=None, config_name='config')
def main(cfg: Config):
  makeup_config(cfg)

  # Setting all the random seeds to the same value.
  # This is important in a distributed training setting.
  # Each rank will get its own set of initial weights.
  # If they don't match up, the gradients will not match either,
  # leading to training that may not converge.
  lightning.seed_everything(cfg.seed)

  if cfg.data.name == 'harmonix':
    dm = HarmonixDataModule(cfg)
  else:
    raise ValueError(f'Unknown dataset: {cfg.data.name}')

  model = AllInOneTrainer(cfg)

  wandb_logger = WandbLogger(
    project='danceformer',
    tags=[
           f'fold{cfg.fold}'
         ] + (
           [cfg.case] if cfg.case else []
         ),
    log_model=False if cfg.debug or cfg.sanity_check or cfg.offline else 'all',
    offline=cfg.debug or cfg.sanity_check or cfg.offline,
  )
  wandb_logger.log_hyperparams(cfg)
  wandb_logger.experiment.define_metric('val/loss', summary='min')

  callbacks = [
    ModelCheckpoint(monitor='val/loss', mode='min'),
    EarlyStopping(
      monitor='val/loss',
      mode='min',
      patience=cfg.early_stopping_patience,
      min_delta=1e-4,
      log_rank_zero_only=True,
      verbose=True
    ),
    LearningRateMonitor(),
  ]
  if cfg.swa_lr > 1e-4:
    callbacks.append(StochasticWeightAveraging(swa_lrs=cfg.swa_lr))

  trainer = Trainer(
    accelerator='cpu' if cfg.debug else 'auto',
    # For some reason, ddp stucks at evaluation...
    # devices=1 if cfg.sanity_check or cfg.debug else 'auto',
    devices=1,
    gradient_clip_val=cfg.gradient_clip,
    logger=wandb_logger,
    callbacks=None if cfg.sanity_check else callbacks,
    check_val_every_n_epoch=cfg.validation_interval_epochs,
    max_epochs=cfg.max_epochs,
    fast_dev_run=cfg.debug and not cfg.sanity_check,
    overfit_batches=cfg.sanity_check_size if cfg.sanity_check else 0,
  )
  if cfg.sanity_check:
    trainer.limit_val_batches = 0

  if trainer.is_global_zero:
    print('=' * 80)
    print('Config')
    print('=' * 80)
    print(OmegaConf.to_yaml(cfg))
    print('=' * 80)

  trainer.fit(
    model=model,
    datamodule=dm,
  )
  print(f'=> Finished training.')

  if trainer.is_global_zero:
    print('=> Running evaluation...')
    evaluate(
      model=model,
      trainer=trainer,
    )


def makeup_config(cfg: Config):
  if cfg.sanity_check:
    cfg.sched = None
    cfg.warmup_epochs = 0
    cfg.weight_decay = 0
    cfg.drop_conv = 0
    cfg.drop_path = 0
    cfg.drop_hidden = 0
    cfg.drop_attention = 0
    cfg.validation_interval_epochs = 50


if __name__ == '__main__':
  main()



import numpy as np
import pandas as pd

from pathlib import Path
from typing import Literal, Union
from numpy.typing import NDArray
from ..datasetbase import DatasetBase
from ...utils import widen_temporal_events
from ...eventconverters import HarmonixConverter
from .....config import Config


class HarmonixDataset(DatasetBase):
  
  def __init__(
    self,
    cfg: Config,
    split: Literal['train', 'val', 'test'],
  ):
    super().__init__(cfg, split)
    
    fold = cfg.fold
    track_ids = sorted(t.stem for t in Path(cfg.data.path_track_dir).glob('*.mp3'))
    folds = np.arange(len(track_ids)) % cfg.total_folds
    
    test_fold = fold
    val_fold = (fold + 1) % cfg.total_folds
    if split == 'train':
      track_ids = [tid for tid, fold in zip(track_ids, folds) if fold not in [test_fold, val_fold]]
    elif split == 'val':
      track_ids = [tid for tid, fold in zip(track_ids, folds) if fold == val_fold]
    elif split == 'test':
      track_ids = [tid for tid, fold in zip(track_ids, folds) if fold == test_fold]
    else:
      raise ValueError(f'Unknown dataset split: {split}')
    
    df = pd.read_csv(cfg.data.path_metadata)
    df['id'] = df['File'].str.split('_').str[0]
    df = df.set_index('id')
    
    self._track_ids = track_ids
    self.feature_dir = (
      Path(cfg.data.path_feature_dir) if cfg.data.demixed
      else Path(cfg.data.path_no_demixed_feature_dir)
    )
    self.df = df
  
  @property
  def track_ids(self):
    return self._track_ids
  
  def load_features(self, track_id: str) -> NDArray:
    return np.load(self.feature_dir / f'{track_id}.npy')
  
  def create_converter(
    self,
    index: int,
    track_id: str,
    num_frames: int,
    start: float,
    end: Union[float, None],
  ) -> HarmonixConverter:
    return HarmonixConverter(
      track_id=track_id,
      total_frames=num_frames,
      sr=self.sample_rate,
      hop=self.hop,
      start=start,
      end=end,
      base_dir=self.cfg.data.path_base_dir,
    )
  
  def __getitem__(self, idx):
    data = super().__getitem__(idx)
    
    track_idx = self.track_ids[idx].split('_')[0]
    row = self.df.loc[track_idx]
    
    # Make a one-hot vector for the true tempo.
    num_bpm_units = 300
    true_bpm_int = row['BPM']
    true_bpm = np.zeros(num_bpm_units, dtype='float32')
    true_bpm[true_bpm_int] = 1.0
    widen_true_bpm = widen_temporal_events(true_bpm, num_neighbors=2)
    
    return dict(
      **data,
      true_bpm=true_bpm,
      widen_true_bpm=widen_true_bpm,
      true_bpm_int=true_bpm_int,
    )



import hydra
import torch

from ..config import Config
from ..spectrogram import extract_spectrograms
from ..demix import demix
from ..helpers import expand_paths, mkpath


@hydra.main(version_base=None, config_name='config')
def main(cfg: Config):
  track_paths = expand_paths([mkpath(cfg.data.path_track_dir) / '*.mp3'])
  demix_dir = mkpath(cfg.data.path_demix_dir) / cfg.demucs_model
  feature_dir = mkpath(cfg.data.path_feature_dir)

  demix_paths = demix(
    paths=track_paths,
    demix_dir=demix_dir,
    device='cuda' if torch.cuda.is_available() else 'cpu'
  )

  spec_paths = extract_spectrograms(
    demix_paths,
    spec_dir=feature_dir,
    multiprocess=True,
  )

  print(f'Preprocessing finished. {len(spec_paths)} spectrograms saved.')


if __name__ == '__main__':
  main()



from lightning import LightningDataModule
from torch.utils.data import DataLoader
from .dataset import HarmonixDataset
from ..collate import collate_fn
from .....config import Config


class HarmonixDataModule(LightningDataModule):
  dataset_train: HarmonixDataset
  dataset_val: HarmonixDataset
  dataset_test: HarmonixDataset
  
  def __init__(self, cfg: Config):
    super().__init__()
    self.cfg = cfg
  
  def setup(self, stage: str):
    if stage == 'fit':
      self.dataset_train = HarmonixDataset(self.cfg, split='train')
      # if self.cfg.debug:
      #   self.dataset_train = Subset(self.dataset_train, range(1))
    
    if stage in ['fit', 'validate']:
      if self.cfg.sanity_check:
        self.dataset_val = self.dataset_train
      else:
        self.dataset_val = HarmonixDataset(self.cfg, split='val')
      # if self.cfg.debug:
      #   self.dataset_val = Subset(self.dataset_val, range(1))
    
    if stage in ['test', 'predict']:
      if self.cfg.sanity_check:
        self.dataset_test = self.dataset_train
      else:
        self.dataset_test = HarmonixDataset(self.cfg, split='test')
      # if self.cfg.debug:
      #   self.dataset_test = Subset(self.dataset_test, range(1))
  
  def train_dataloader(self):
    return DataLoader(
      self.dataset_train,
      batch_size=self.cfg.batch_size,
      shuffle=True,
      num_workers=0 if self.cfg.debug else 2,
      collate_fn=collate_fn,
    )
  
  def val_dataloader(self):
    return DataLoader(
      self.dataset_val,
      batch_size=1,
      shuffle=False,
      num_workers=0 if self.cfg.debug else 1,
      collate_fn=collate_fn,
    )
  
  def test_dataloader(self):
    return DataLoader(
      self.dataset_test,
      batch_size=1,
      shuffle=False,
      num_workers=0 if self.cfg.debug else 1,
      collate_fn=collate_fn,
    )
  
  def predict_dataloader(self):
    return self.test_dataloader()



import pandas as pd
from abc import ABC, abstractmethod
from pathlib import Path
from .eventconverters import BeatConverter, DownbeatConverter, SectionConverter
from ....typings import PathLike
from ....config import HARMONIX_LABELS


class DatasetConverter(ABC):

  @property
  @abstractmethod
  def beat(self) -> BeatConverter:
    pass

  @property
  @abstractmethod
  def downbeat(self) -> DownbeatConverter:
    pass

  @property
  @abstractmethod
  def section(self) -> SectionConverter:
    pass


class HarmonixConverter(DatasetConverter):

  def __init__(
    self,
    track_id: str,
    total_frames: int = None,
    *,
    sr: int,
    hop: int,
    start: float = None,
    end: float = None,
    base_dir: PathLike = './data/harmonix/',
  ):
    self.base_dir = Path(base_dir)
    self.track_id = track_id
    self.total_frames = total_frames
    self.sr = sr
    self.hop = hop
    self.start = start
    self.end = end

    self.df_beat = pd.read_csv(
      self.base_dir / 'beats' / f'{self.track_id}.txt',
      names=['time', 'count'],
      delimiter='\t'
    )
    self.df_downbeat = self.df_beat[self.df_beat['count'] == 1]
    self.df_section = pd.read_csv(
      self.base_dir / 'segments' / f'{self.track_id}.txt',
      names=['start', 'name'],
      delimiter='\t',
    )

    section_times = self.df_section['start'].values
    section_labels = self.df_section['name'].values.tolist()
    # if section_times[0] > 0.0:
    section_labels = ['start'] + section_labels

    self._beat = BeatConverter(
      self.df_beat['time'].values,
      segment_frames=total_frames,
      sr=sr,
      hop=hop,
      start=start,
      end=end,
    )
    self._downbeat = DownbeatConverter(
      self.df_downbeat['time'].values,
      segment_frames=total_frames,
      sr=sr,
      hop=hop,
      start=start,
      end=end,
    )
    self._section = SectionConverter(
      times=section_times,
      section_labels=section_labels,
      label_vocab=HARMONIX_LABELS,
      beat_times=self.beat.times,
      segment_frames=total_frames,
      sr=sr,
      hop=hop,
      start=start,
      end=end,
    )

  @property
  def beat(self) -> BeatConverter:
    return self._beat

  @property
  def downbeat(self) -> DownbeatConverter:
    return self._downbeat

  @property
  def section(self) -> SectionConverter:
    return self._section



import numpy as np

import torch
import librosa
import demucs.separate

from functools import partial
from multiprocessing import Pool
from typing import Union, List, Tuple
from tqdm import tqdm
from numpy.typing import NDArray
from .typings import AnalysisResult, PathLike, Segment
from .utils import mkpath


def sonify(
  results: Union[AnalysisResult, List[AnalysisResult]],
  out_dir: PathLike = None,
  multiprocess: bool = True,
) -> Union[Tuple[NDArray, float], List[Tuple[NDArray, float]]]:
  return_list = True
  if not isinstance(results, list):
    return_list = False
    results = [results]

  sonif_fn = partial(_sonify, out_dir=out_dir)
  if multiprocess:
    pool = Pool()
    iterator = pool.imap_unordered(sonif_fn, results)
  else:
    iterator = map(sonif_fn, results)

  sonifs = [sonif for sonif in tqdm(iterator, desc='Sonifying results', total=len(results))]

  if multiprocess:
    pool.close()
    pool.join()

  if not return_list:
    return sonifs[0]
  return sonifs


def _sonify(
  result: AnalysisResult,
  out_dir: PathLike = None,
) -> Tuple[NDArray, float]:
  sr = 44100
  y = demucs.separate.load_track(result.path, 2, sr).numpy()
  # y, sr = librosa.load(result.path, sr=None, mono=False)

  length = y.shape[-1]
  metronome = _sonify_metronome(result, length, sr)
  boundaries = _sonify_boundaries(result.segments, length, sr)

  mixed = y + metronome + boundaries
  mixed = np.clip(mixed, -1, 1)

  if out_dir is not None:
    out_dir = mkpath(out_dir)
    out_dir.mkdir(exist_ok=True, parents=True)

    demucs.separate.save_audio(
      wav=torch.from_numpy(mixed),
      path=out_dir / f'{result.path.stem}.sonif{result.path.suffix}',
      samplerate=sr,
    )

  return mixed, sr


def _sonify_metronome(
  result: AnalysisResult,
  length: int,
  sr: float = 44100,
):
  # Exclude downbeats from beats.
  downbeats = np.asarray(result.downbeats)
  beats = np.asarray(result.beats)
  dists = np.abs(downbeats[:, np.newaxis] - beats).min(axis=0)
  beats = beats[dists > 0.03]

  clicks_beat = librosa.clicks(
    times=beats,
    sr=sr,
    click_freq=1500,
    click_duration=0.1,
    length=length,
  )
  clicks_downbeat = librosa.clicks(
    times=downbeats,
    sr=sr,
    click_freq=3000,
    click_duration=0.1,
    length=length,
  )

  return clicks_beat + clicks_downbeat


def _sonify_boundaries(
  segments: List[Segment],
  length: int,
  sr: float = 44100,
  riser_freq_start: float = 40.0,
  riser_freq_end: float = 4000.0,
  riser_duration: float = 4.0,
  num_clicks: int = 25,
):
  click_offsets = -np.geomspace(riser_duration, 0.1, num_clicks, endpoint=False)
  boundaries = [segment.start for segment in segments if segment.label not in ['start', 'end']]
  click_times = click_offsets[:, np.newaxis] + boundaries
  click_times = click_times.T.flatten()
  click_freqs = np.geomspace(riser_freq_start, riser_freq_end, num_clicks)
  click_freqs = np.tile(click_freqs, len(boundaries))

  click_freqs = click_freqs[click_times > 0]
  click_times = click_times[click_times > 0]

  y = np.zeros((length,), dtype='float32')
  for t, f in zip(click_times, click_freqs):
    click = _synthesize_click(sr, f)

    click_start = int(t * sr)
    click_end = click_start + len(click)
    if click_end > length:
      click = click[:length - click_start]

    y[click_start:click_start + len(click)] += click

  for segment in segments:
    if segment.label in ['start', 'end']:
      continue

    drop = _synthesize_drop(sr)

    drop_start = int(segment.start * sr)
    drop_end = drop_start + len(drop)
    if drop_end > length:
      drop = drop[:length - drop_start]

    y[drop_start:drop_start + len(drop)] += drop

  return y


def _synthesize_click(
  sr: float = 44100,
  click_freq: float = 1000.0,
  click_duration: float = 0.1,
):
  angular_freq = 2 * np.pi * click_freq / float(sr)
  click = np.logspace(0, -10, num=int(sr * click_duration), base=2.0)
  click *= np.sin(angular_freq * np.arange(len(click)))
  return click


def _synthesize_drop(
  sr: float = 44100,
  drop_freq_start: float = 4000.0,
  drop_freq_end: float = 40.0,
  drop_duration: float = 0.5,
):
  freqs = np.geomspace(drop_freq_start, drop_freq_end, int(sr * drop_duration))
  drop = np.sin(2 * np.pi * freqs.cumsum() / sr)
  drop *= np.logspace(0, -4, num=int(sr * drop_duration), base=2.0)
  return drop



import numpy as np
import librosa
from abc import ABC
from typing import List
from numpy.typing import NDArray


class EventConverter(ABC):
  def __init__(
    self,
    times: NDArray[np.float64],
    segment_frames: int = None,
    *,
    sr: int,
    hop: int,
    start: float = None,
    end: float = None,
    # encode: bool = False,
  ):
    self.times = times
    self.segment_frames = segment_frames
    self.sr = sr
    self.hop = hop
    self.start = start
    self.end = end
  
  def get_start_end_frames(self):
    start_frame = librosa.time_to_frames(self.start, sr=self.sr, hop_length=self.hop).item() if self.start else 0
    start_frame = int(start_frame)
    if self.segment_frames:
      if self.end is not None:
        raise ValueError('total_frames and end cannot be used together')
      segment_frames = self.segment_frames
    else:
      assert self.end is not None, 'either total_frames or end should be given'
      segment_seconds = self.end - self.start
      segment_frames = round(segment_seconds * self.sr / self.hop)
    
    end_frame = start_frame + segment_frames
    
    return start_frame, end_frame
  
  def frames(
    self,
    reset_index: bool = False,
  ):
    frames = librosa.time_to_frames(self.times, sr=self.sr, hop_length=self.hop)
    start_frame, end_frame = self.get_start_end_frames()
    frames = frames[(start_frame <= frames) & (frames < end_frame)]
    
    if reset_index and self.start is not None:
      frames -= start_frame
    
    return frames
  
  def samples(
    self,
    reset_index: bool = False,
  ):
    samples = librosa.time_to_samples(self.times, sr=self.sr)
    
    if self.start is not None:
      start_sample = librosa.time_to_samples(self.start, sr=self.sr)
      samples = samples[start_sample <= samples]
    
    if self.end is not None:
      end_sample = librosa.time_to_samples(self.end, sr=self.sr)
      samples = samples[samples < end_sample]
    
    if reset_index and self.start is not None:
      samples -= start_sample
    
    return samples
  
  def of_frames(
    self,
    encode: bool = False,
  ):
    start_frame, end_frame = self.get_start_end_frames()
    num_frames = end_frame - start_frame
    
    if encode:
      event_frames = self.frames(reset_index=True)
      out = np.zeros(num_frames, dtype='float32')
      out[event_frames] = 1.
    else:
      event_frames = self.frames(reset_index=False)
      i_start, i_end = np.searchsorted(event_frames, [start_frame, end_frame])
      out = event_frames[i_start:i_end]
    
    return out


class BeatConverter(EventConverter):
  pass


class DownbeatConverter(EventConverter):
  pass


class SectionConverter(EventConverter):
  
  def __init__(
    self,
    times: NDArray[np.float64],
    section_labels: List[str],
    label_vocab: List[str],
    beat_times: NDArray[np.float64],
    segment_frames: int = None,
    *,
    sr: int,
    hop: int,
    start: float = None,
    end: float = None,
  ):
    assert len(times) + 1 == len(section_labels), 'section_labels should be one longer than section boundaries'
    super().__init__(
      times=times,
      segment_frames=segment_frames,
      sr=sr,
      hop=hop,
      start=start,
      end=end,
    )
    self.section_labels = section_labels
    self.beat_times = beat_times
    
    # Check if all_labels are unique.
    assert len(label_vocab) == len(set(label_vocab)), 'all_labels should be unique'
    self.label_vocab = label_vocab
    self.label_map = {label: i for i, label in enumerate(label_vocab)}
  
  def of_beats(self, beat_times: NDArray[np.float64] = None):
    beat_times = beat_times or self.beat_times
    names = np.array(self.section_labels)
    indices = np.searchsorted(self.times, beat_times, side='right')
    beat_sections = names[indices]
    return beat_sections
  
  def of_frames(
    self,
    encode: bool = False,
    return_labels: bool = True,
  ):
    if return_labels:
      boundary_frames = librosa.time_to_frames(self.times, sr=self.sr, hop_length=self.hop)
      start_frame, end_frame = self.get_start_end_frames()
      frame_indices = np.arange(start_frame, end_frame)
      indices = np.searchsorted(boundary_frames, frame_indices, side='right')
      labels = np.array(self.section_labels)
      if encode:
        labels = np.array([self.label_map[l] for l in labels])
      frame_labels = labels[indices]
      
      return frame_labels
    else:
      return super().of_frames(encode)
  
  @property
  def labels(self):
    assert 'end' in self.section_labels[-1]
    label_names = self.section_labels
    return np.array([self.label_map[l] for l in label_names])



import numpy as np

from abc import ABC, abstractmethod
from typing import Literal, List, Union
from numpy.typing import NDArray
from torch.utils.data import Dataset
from ..utils import widen_temporal_events
from ..eventconverters import DatasetConverter
from ....config import Config


class DatasetBase(Dataset, ABC):

  def __init__(
    self,
    cfg: Config,
    split: Literal['train', 'val', 'test', 'unseen'],
  ):
    if split not in ['train', 'val', 'test', 'unseen']:
      raise ValueError(f'Unknown dataset split: {split}')

    self.cfg = cfg
    self.split = split
    self.sample_rate = cfg.sample_rate
    self.hop = cfg.hop_size
    self.segment_size = cfg.segment_size if split == 'train' else None

  @abstractmethod
  def load_features(self, track_id: str) -> NDArray:
    pass

  @property
  @abstractmethod
  def track_ids(self) -> List[str]:
    pass

  @abstractmethod
  def create_converter(
    self,
    index: int,
    track_id: str,
    num_frames: int,
    start: float,
    end: Union[float, None],
  ) -> DatasetConverter:
    pass

  def __len__(self):
    return len(self.track_ids)

  def __getitem__(self, idx):
    track_id = self.track_ids[idx]
    spec_full = self.load_features(track_id)

    should_segment = self.segment_size is not None
    if should_segment:
      duration = spec_full.shape[1] / self.cfg.fps
      num_frames = int(self.segment_size * self.cfg.fps)
      start = (
        0 if self.cfg.sanity_check
        else np.random.uniform(0, max(0, duration - self.segment_size))
      )
      end = None
    else:
      num_frames = spec_full.shape[1]
      start = None
      end = None

    st = self.create_converter(idx, track_id, num_frames, start, end)
    start_frame, end_frame = st.beat.get_start_end_frames()
    spec = spec_full[:, start_frame:end_frame, :]

    # Normalization should be done here, but it seems madmom does it for us.

    true_beat = st.beat.of_frames(encode=True)
    true_downbeat = st.downbeat.of_frames(encode=True)
    true_section = st.section.of_frames(encode=True, return_labels=False)
    true_function = st.section.of_frames(encode=True, return_labels=True)
    true_function_list = st.section.labels

    # widen the temporal activation
    # region around the annotations to include two adjacent temporal
    # frames on either side of each quantised beat location and
    # weight them with a value of 0.5 during training.
    widen_true_beat = widen_temporal_events(true_beat, num_neighbors=1)
    widen_true_downbeat = widen_temporal_events(true_downbeat, num_neighbors=1)
    widen_true_section = widen_temporal_events(true_section, num_neighbors=2)

    true_beat_times = st.beat.times
    true_downbeat_times = st.downbeat.times
    true_section_times = st.section.times

    if should_segment:
      end_time = start + self.segment_size
      true_beat_times = true_beat_times[(start <= true_beat_times) & (true_beat_times < end_time)]
      true_downbeat_times = true_downbeat_times[(start <= true_downbeat_times) & (true_downbeat_times < end_time)]
      true_section_times = true_section_times[(start <= true_section_times) & (true_section_times < end_time)]

      true_beat_times -= start
      true_downbeat_times -= start
      true_section_times -= start

    return dict(
      track_key=track_id,
      spec=spec,

      true_beat=true_beat,
      true_downbeat=true_downbeat,
      true_section=true_section,
      true_function=true_function,

      widen_true_beat=widen_true_beat,
      widen_true_downbeat=widen_true_downbeat,
      widen_true_section=widen_true_section,

      true_beat_times=true_beat_times.tolist(),
      true_downbeat_times=true_downbeat_times.tolist(),
      true_section_times=true_section_times.tolist(),
      true_function_list=true_function_list.tolist(),
    )
